r"""

.. todo::
    merge docs with those in :code:`realgas.partial_molar_properties`.
    There, the definitions of residual properties are displayed and here wee need only
    write simplified forms for the specific equation of state.

Theory
------

The second order virial equation of state is :cite:`Perry`

.. math::
    Z = 1 + B\rho = 1 + \frac{BP}{RT}
    :label: z_virial

Where the composition dependency of :math:`B` is given by the *exact* mixing rule

.. math::
    B = \sum_i \sum_j y_i y_j B_{ij}
    :label: B_mix_expr

where :math:`B_{ij}=B_{ji}`, and :math:`B_{ii}` and :math:`B_{jj}` are virial coefficients for the pure species

In this package, the useful correlation

.. math::
    \frac{BP_\text{c}}{RT_\text{c}} = B^0 + \omega B^1

or

.. math::
    B = \frac{RT_\text{c}}{P_\text{c}}\left(B^0 + \omega B^1\right)
    :label: B_expr

So that, combining Equations :eq:`z_virial` and :eq:`B_expr`,
the compressibility can be calculated from dimensionless quantities as

.. math::
    Z = 1 + \left(B^0 + \omega B^1\right)\frac{P_\text{r}}{T_\text{r}}
    :label: z_dimensionless

where :cite:`Smith2005`

.. math::
    B^0 = 0.083 - \frac{0.422}{T_\text{r}^{1.6}}
    :label: B0_expr

.. math::
    B^1 = 0.139 - \frac{0.172}{T_\text{r}^{4.2}}
    :label: B1_expr

so that

so that the following derivatives can be computed as

.. math::
    \frac{dB^0}{dT_\text{r}} = \frac{0.675}{T_\text{r}^{2.6}}
    :label: dB0dTr

.. math::
    \frac{dB^1}{dT_\text{r}} = \frac{0.722}{T_\text{r}^{5.2}}
    :label: dB1dTr

Which allow the :math:`H^\text{R}`, :math:`S^\text{R}`, and :math:`G^\text{R}` to be readily computed as follows :cite:`Perry`

.. math::
    \frac{G^\text{R}}{RT} = \left(B^0 + \omega B^1\right)\frac{P_\text{r}}{T_\text{r}}
    :label: G_R_RT_virial

.. math::
    \frac{H^\mathrm{R}}{RT} = P_\mathrm{r}\left[
        \frac{B^0}{T_\mathrm{r}} - \frac{\mathrm{d} B^0}{\mathrm{d}T_\mathrm{r}} + \omega\left(\frac{B^1}{T_\mathrm{r}} - \frac{\mathrm{d}B^1}{\mathrm{d}T_\mathrm{r}}\right)
    \right]
    :label: H_R_RT_virial

.. math::
    \frac{S^\text{R}}{R} = -P_\text{r}\left(\frac{d B^0}{d T_\text{r}} - \omega\frac{d B^1}{d T_\text{r}}\right)
    :label: S_R_R_virial


The cross coefficients are calculated as

.. math::
    B_{ij} = \frac{RT_{\text{c}ij}}{P_{\text{c}ij}}\left(B^0 + \omega_{ij}B^1\right)
    :label: B_ij_expr

so that the cross derivatives can be computed as

.. math::
    \frac{dB_{ij}}{dT} = \frac{RT_{\text{c}ij}}{P_{\text{c}ij}}\left(\frac{dB^0}{d T} + \omega_{ij}\frac{dB^1}{dT}\right)

    \frac{dB_{ij}}{dT} = \frac{R}{P_{\text{c}ij}}\left(\frac{dB^0}{d T_{\text{r}ij}} + \omega_{ij}\frac{dB^1}{dT_{\text{r}ij}}\right)

or, equivalently,

.. math::
    \frac{dB_{ij}}{dT_{\text{r}ij}} = \frac{RT_{\text{c}ij}}{P_{\text{c}ij}}\left(\frac{dB^0}{d T_{\text{r}ij}} + \omega_{ij}\frac{dB^1}{dT_{\text{r}ij}}\right)
    :label: dB_dTrij


Fugacity Coefficients
---------------------

The Fugacity coefficients are calculated as

.. math::
    \ln\hat{\phi}_i=\left(\frac{\partial (nG^\text{R}/R/T)}{\partial n_i}\right)_{P,T,n_j}

For the virial equation of state, this becomes :cite:`VanNess1982`

.. math::
    \ln\hat{\phi}_k = \frac{P}{RT}\left[B_{kk} + \frac{1}{2}\sum_i\sum_jy_iy_j\left(2\delta_{ik} - \delta_{ij}\right)\right]
    :label: ln_phi_i_virial

where *both* :math:`i` and :math:`j` indices run over all species

.. math::
    \delta_{ik} = 2B_{ik} - B_{ii} - B_{kk} = \delta_{ki}
    :label: d_ik

and

.. math::
    \delta_{ii} = 0

Residual Partial Molar Properties
---------------------------------
The partial molar residual free energy of component *k* is

.. math::
    \frac{\bar{G}^\text{R}_k}{RT} = \ln{\hat{\phi}_k} = \frac{P}{RT}\left[B_{kk} + \frac{1}{2}\sum_i\sum_jy_iy_j\left(2\delta_{ik} - \delta_{ij}\right)\right]
    :label: barGiR


The partial molar residual enthalpy of component *k* is


.. math::
    \begin{align}
        \frac{\bar{H}^\mathrm{R}_k}{RT} &= -T \left(\frac{\partial \ln\hat{\phi}_k}{\partial T}\right)_{P,y} \\
                             &= - \frac{T}{T_c} \left(\frac{\partial \ln\hat{\phi}_k}{\partial T_r}\right)_{P,y} \\
                             &= - \frac{T}{T_c}\left\{
                                \frac{P}{RT}\left[
                                    \frac{\partial B_{kk}}{\partial T_r}
                                     + \frac{1}{2}\sum_i\sum_j y_iy_j\left(
                                                    2\frac{\partial \delta_{ik}}{\partial T_r}
                                                    - \frac{\partial \delta_{ij}}{\partial T_r}
                                                    \right)
                                    \right]
                                - \frac{T_c\ln\hat{\phi}_i}{T}
                             \right\}
    \end{align}

where :math:`\frac{\partial \delta_{ij}}{\partial T_r}` is given by :eq:`eq_d_dij`
so that we obtain

.. math::
    \frac{\bar{H}^\mathrm{R}_k}{RT} = - \frac{P}{RT_\text{c}}\left[
                                    \frac{\partial B_{kk}}{\partial T_r}
                                     + \frac{1}{2}\sum_i\sum_j y_iy_j\left(
                                                    2\frac{\partial \delta_{ik}}{\partial T_r}
                                                    - \frac{\partial \delta_{ij}}{\partial T_r}
                                                    \right)
                                    \right]
                                + \ln\hat{\phi}_i
    :label: barHiR

Since

.. math::
    G^\mathrm{R} = H^\mathrm{R} - T S^\mathrm{R}

In terms of partial molar properties, then

.. math::
    \begin{align}
        \bar{S}_i^\mathrm{R} &= \frac{\bar{H}_i^\mathrm{R} - \bar{G}_i^\mathrm{R}}{T} \\
        \frac{\bar{S}_i^\mathrm{R}}{R} &= \frac{\bar{H}_i^\mathrm{R}}{RT} - \frac{\bar{G}_i^\mathrm{R}}{RT} \\
    \end{align}

By comparing Equation :eq:`barGiR` and :eq:`barHiR` it is observed that

.. math::
    \frac{\bar{S}_k^\mathrm{R}}{R} = - \frac{P}{RT_\text{c}}\left[
                                    \frac{\partial B_{kk}}{\partial T_r}
                                     + \frac{1}{2}\sum_i\sum_j y_iy_j\left(
                                                    2\frac{\partial \delta_{ik}}{\partial T_r}
                                                    - \frac{\partial \delta_{ij}}{\partial T_r}
                                                    \right)
                                    \right]
    :label: barSiR

where :math:`\frac{\partial \delta_{ij}}{\partial T_r}` is given by :eq:`eq_d_dij`



The partial molar residual volume of component *i* is calculated as

.. math::
    \begin{align}
        \frac{\bar{V}_k^\mathrm{R}}{RT} &= \left(\frac{\partial \ln\hat{\phi}_i}{\partial P}\right)_{T,y}\\
            &= \frac{\partial}{\partial P}\left\{\frac{P}{RT}\left[B_{kk} + \frac{1}{2}\sum_i\sum_jy_iy_j\left(2\delta_{ik} - \delta_{ij}\right)\right]\right\}
    \end{align}

which simplifies to

.. math::
    \frac{\bar{V}_k^\mathrm{R}}{RT} = \frac{1}{RT}\left[B_{kk} + \frac{1}{2}\sum_i\sum_jy_iy_j\left(2\delta_{ik} - \delta_{ij}\right)\right]
    :label: barViR

from which we obtain the intuitive result of

.. math::
    \bar{V}_k^\mathrm{R} = B_{kk} + \frac{1}{2}\sum_i\sum_jy_iy_j\left(2\delta_{ik} - \delta_{ij}\right)


"""

import typing

import matplotlib.pyplot as plt
import numpy as np
from chem_util.chem_constants import gas_constant as R

from ..critical_constants import CriticalConstants
from ..input import RealMixture


class Virial:
    """

    :param pow: function for computing power, defaults to numpy.power
    :type pow: callable, optional
    :param exp: function for computing logarithm, defaults to numpy.exp
    :type exp: callable, optional
    """

    def __init__(self, pow: callable = np.power, exp: callable = np.exp):
        self.pow = pow
        self.exp = exp

    def B0_expr(self, T_r):
        """
        :param T_r: Reduced temperature
        :returns: Equation :eq:`B0_expr`
        """
        return 0.083 - 0.422 * self.pow(T_r, -1.6)

    def B1_expr(self, T_r):
        """

        :param T_r: reduced temperature
        :return: Equation eq:`B1_expr`
        """
        return 0.139 - 0.172 * self.pow(T_r, -4.2)

    def d_B0_d_Tr_expr(self, T_r):
        """
        :param T_r: reduced temperature
        :returns: Equation :eq:`dB0dTr`
        """
        return 0.6752 * self.pow(T_r, -2.6)

    def d_B1_d_Tr_expr(self, T_r):
        """
        :param T_r: reduced temperature
        :returns: Equation :eq:`dB1dTr`
        """
        return 0.7224 * self.pow(T_r, -5.2)

    def B_expr(self, T_r, w, T_c, P_c):
        """

        :param T_r: reduced temperature
        :param w: accentric factor
        :param T_c: critical temperautre [K]
        :param P_c: critical pressure [Pa]
        :return: Equation :eq:`B_expr`
        """
        return R * T_c / P_c * (
                self.B0_expr(T_r) + w * self.B1_expr(T_r)
        )

    def ln_hat_phi_k_expr(self, *args):
        raise NotImplementedError

    def hat_phi_i_expr(self, *args):
        r"""expression for fugacity coefficient
        :returns: :math:`\exp\left({\ln{\hat{\phi}_i}}\right)`
        """
        return self.exp(self.ln_hat_phi_k_expr(*args))


class SecondVirial(CriticalConstants, Virial):
    """Virial equation of state for one component. See :cite:`Perry,Smith2005`

    """

    def __init__(self, dippr_no: str = None, compound_name: str = None, cas_number: str = None,
                 pow: callable = np.power, **kwargs):
        """

        :param kwargs: for customized critica constants
        """
        CriticalConstants.__init__(self, dippr_no, compound_name, cas_number, **kwargs)
        Virial.__init__(self, pow=pow)
        self.name = 'secondvirial'

    def H_R_RT_expr(self, P, T):
        """

        :param P: pressure in Pa
        :param T: temperature in K
        :return: Equation :eq:`H_R_RT_virial`
        """
        T_r = T / self.T_c
        return P / self.P_c * (
                self.B0_expr(T_r) / T_r - self.d_B0_d_Tr_expr(T_r) + self.w * (
                self.B1_expr(T_r) / T_r - self.d_B1_d_Tr_expr(T_r))
        )

    def G_R_RT_expr(self, P, T):
        """

        :param P: pressure in Pa
        :param T: Temperautre in K
        :return: Equation :eq:`G_R_RT_virial`
        """
        T_r = T / self.T_c
        P_r = P / self.P_c
        return (self.B0_expr(T_r) + self.w * self.B1_expr(T_r)) * P_r / T_r

    def S_R_R_expr(self, P, T):
        """

        :param P: pressure in Pa
        :param T: temperature in K
        :return: Equation :eq:`S_R_R_virial`
        """
        T_r = T / self.T_c
        P_r = P / self.P_c
        return -P_r * (self.d_B0_d_Tr_expr(T_r) + self.w * self.d_B1_d_Tr_expr(T_r))

    def ln_hat_phi_k_expr(self, P, T):
        r"""logarithm of fugacity coefficient

        .. note::
            single-component version

        In this case, Equation :eq:`ln_phi_i_virial` simplifies to

        .. math::
            \ln\hat{\phi}_i = \frac{PB}{RT}

        :param P: pressure in Pa
        :type P: float
        :param T: temperature in K
        :type T: float
        """
        T_r = T / self.T_c
        return P * self.B_expr(T_r, self.w, self.T_c, self.P_c) / R / T

    def calc_Z_from_units(self, P, T):
        """

        :param P: pressure in Pa
        :param T: temperature in K
        :return: Equation :eq:`z_virial`
        """
        return 1. + self.B_expr(T / self.T_c, self.w, self.T_c, self.P_c) * P / R / T

    def calc_Z_from_dimensionless(self, P_r, T_r):
        """

        :param P_r: reduced pressure, dimensionless
        :param T_r: reduced temperature, dimensionless
        :return: Equation :eq:`z_dimensionless`
        """
        return 1 + (self.B0_expr(T_r) + self.w * self.B1_expr(T_r)) * P_r / T_r

    def plot_Z_vs_P(self, T, P_min, P_max, symbol='o', ax=None, **kwargs):
        """Plot compressibility as a function of pressure

        :param T: temperature [K]
        :type T: float
        :param P_min: minimum pressure for plotting [Pa]
        :type P_min: float
        :param P_max: maximum pressure for plotting [Pa]
        :type P_max: float
        :param phase: phase type (liquid or vapor), defaults to vapor
        :type phase: str
        :param symbol: marker symbol, defaults to 'o'
        :type symbol: str
        :param ax: matplotlib axes for plotting, defaults to None
        :type ax: plt.axis
        :param kwargs: keyword arguments for plotting
        """
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.set_ylabel('$Z$')
            ax.set_xlabel('$P$ [Pa]')

        P = np.linspace(P_min, P_max)
        Z = [self.calc_Z_from_dimensionless(pressure / self.P_c, T / self.T_c) for pressure in P]
        ax.plot(P, Z, symbol, markerfacecolor='None', **kwargs)


class MixingRule(Virial):
    r"""Van der Walls mixing rule

    combining rules from :cite:`Prausnitz1986`


    .. math::
        \omega_{ij} = \frac{\omega_i + \omega_j}{2}
        :label: omega_combine

    .. math::
        T_{\text{c}ij} = \sqrt{T_{\text{c}i}T_{\text{c}j}}(1-k_{ij})
        :label: Tc_combine

    .. math::
        P_{\text{c}ij} = \frac{Z_{\text{c}ij}RT_{\text{c}ij}}{V_{\text{c}ij}}
        :label: Pc_combine

    .. math::
        Z_{\text{c}ij} = \frac{Z_{\text{c}i} + Z_{\text{c}j}}{2}
        :label: Zc_combine

    .. math::
        V_{\text{c}ij} = \left(\frac{V_{\text{c}i}^{1/3} + V_{\text{c}j}^{1/3}}{2}\right)^3
        :label: vc_combine


    """

    def __init__(self, pow: callable = np.power, exp: callable = np.exp):
        Virial.__init__(self, pow, exp)

    def w_ij_rule(self, w_i, w_j):
        r"""

        :param w_i: accentric factor of component *i*
        :param w_j: accentric factor of component *j*
        :return: Equation :eq:`omega_combine`
        """
        return (w_i + w_j) / 2.

    def T_cij_rule(self, T_ci, T_cj, k_ij):
        """

        :param T_ci: critical temperature of component i [K]
        :param T_cj: ** j [K]
        :param k_ij: k_ij parameter
        :return: Equation :eq:`Tc_combine`
        """
        return pow(T_ci * T_cj, 0.5) * (1. - k_ij)

    def P_cij_rule(self, Z_ci, V_ci, T_ci, Z_cj, V_cj, T_cj, k_ij):
        """

        :return: Equation :eq:`Pc_combine`
        """
        Z_cij = self.Z_cij_rule(Z_ci, Z_cj)
        T_cij = self.T_cij_rule(T_ci, T_cj, k_ij=k_ij)
        V_cij = self.V_cij_rule(V_ci, V_cj)
        return Z_cij * R * T_cij / V_cij

    def Z_cij_rule(self, Z_ci, Z_cj):
        """

        :param Z_ci: critical compressibility factor of component *i*
        :param Z_cj: critical compressibility factor of component *j*
        :return: Equation :eq:`Zc_combine`
        """
        return (Z_ci + Z_cj) / 2.

    def V_cij_rule(self, V_ci, V_cj):
        """

        :param V_ci: critical molar volume of component *i* [m**3/mol]
        :param V_cj: critical molar volume of component *j* [m**3/mol]
        :return: Equation eq:`Vc_combine`
        """
        return self.pow(
            (self.pow(V_ci, 1 / 3) + self.pow(V_cj, 1 / 3)) / 2., 3
        )


class SecondVirialMixture(RealMixture, MixingRule):
    r"""Second virial with mixing rule from :class:`.MixingRule`

    .. note::
        can only input both custom critical properties or both from DIPPR--cant have mixed at the moment

    :param pow: function to calculate power, defaults to numpy.power
    :param exp: function to calculate exponential,d efaults to numpy.exp
    :param kwargs: key-word arguments to pass to :class:`.RealMixture`

    """

    def __init__(self,
                 pow: callable = np.power, exp: callable = np.exp,
                 **kwargs):
        MixingRule.__init__(self, pow, exp)
        RealMixture.__init__(self, **kwargs)
        RealMixture.setup(self)

        for i in range(self.num_components):
            I = CriticalConstants(**self.get_point_input(i))
            self.set_point_input(i, **I.__dict__)
        self.check_params()

    def get_w_Tc_Pc(self, i: int, j=None):
        """Returns critical constants for calculation based off of whetner i = j or not

        :returns: (:math:`w`, :math:`T_c`, :math:`P_c`)
        :rtype: tuple
        """
        if j is None or i == j:
            return self.ws[i], self.T_cs[i], self.P_cs[i]

        return (
            self.w_ij_rule(self.ws[i], self.ws[j]),
            self.T_cij_rule(self.T_cs[i], self.T_cs[j], k_ij=self.k_ij[i][j]),
            self.P_cij_rule(
                self.Z_cs[i], self.V_cs[i], self.T_cs[i],
                self.Z_cs[j], self.V_cs[j], self.T_cs[j],
                k_ij=self.k_ij[i][j]
            )
        )

    def B_ij_expr(self, i: int, j: int, T):
        r"""

        :param i: index of first component
        :param j: index of second component
        :param T: temperature [K]
        :returns: Equation :eq:`B_ij_expr`
        """
        w, T_c, P_c = self.get_w_Tc_Pc(i, j)
        T_r = T / T_c
        return self.B_expr(T_r, w, T_c, P_c)

    def B_mix_expr(self, y_k: typing.List[typing.Union[float, typing.Any]], T):
        """

        :param y_k: mole fractions of each component :math:`k`
        :param T: temperature in K
        :returns: Equation :eq:`B_mix_expr`
        """
        return sum(
            y_k[i] * y_k[j] * self.B_ij_expr(i, j, T) for i in range(self.num_components) for j in
            range(self.num_components)
        )

    def calc_Z_from_units(self, y_k: typing.List, P, T):
        """

        :param y_k: mole fractions of each component :math:`k`
        :param P: pressure in Pa
        :param T: temperature in K
        :return: Equation :eq:`z_virial`
        """
        return 1. + self.B_mix_expr(y_k, T) * P / R / T

    def d_ik_expr(self, i: int, k: int, T):
        """

        :param i: index of component *i*
        :param k: index of component *k*
        :param T: temperature [K]
        :return: Equation :eq:`d_ik`
        """
        if i == k:
            return 0.

        return 2. * self.B_ij_expr(i, k, T) - self.B_ij_expr(i, i, T) - self.B_ij_expr(k, k, T)

    def ln_hat_phi_k_expr(self, k: int, ys: typing.List[typing.Union[float, typing.Any]], P, T):
        r"""logarithm of fugacity coefficient

        :param k: index of component *k*
        :param ys: mole fractions
        :param P: pressure in Pa
        :param T: temperature in K
        :returns: Equation :eq:`ln_phi_i_virial`
        """
        return P / R / T * (
                self.B_ij_expr(k, k, T)
                + 1. / 2. * sum(
            ys[i] * ys[j] * (2. * self.d_ik_expr(i, k, T) - self.d_ik_expr(i, j, T))
            for i in range(self.num_components) for j in range(self.num_components)
        )
        )

    def fugacity_i_expr(self, cas_i: str, ys: typing.List[typing.Union[float, typing.Any]], P, T):
        r"""Fugacity of component i in mixture :math:`f_i=\hat{\phi}_i y_i P`

        :param cas_i: cas number for component of interest
        :param ys: mole fractions
        :param P: pressure in Pa
        :param T: temperature in K
        """
        assert cas_i in self.cas_numbers, 'Cas number not found!'
        i = self.cas_numbers.index(cas_i)
        return self.hat_phi_i_expr(i, ys, P, T) * ys[i] * P

    def bar_GiR_RT(self, cas_k: str, ys: typing.List[typing.Union[float, typing.Any]], P, T):
        r"""Dimensionless residual partial molar free energy of component :math:`i`

        :param cas_k: cas number of component k
        :param ys: mole fractions
        :param P: pressure in Pa
        :param T: temperature in K
        :returns: Equation :eq:`barGiR`
        """
        assert cas_k in self.cas_numbers, 'Cas number not found'
        k = self.cas_numbers.index(cas_k)
        return self.ln_hat_phi_k_expr(k, ys, P, T)

    def d_Bij_d_Trij(self, i: int, j: int, T):
        """

        :param i: index for component *i*
        :param j: index for componen t *j*
        :param T: temperature in K
        :return: Equation :eq:`dB_dTrij`
        """
        w, T_c, P_c = self.get_w_Tc_Pc(i, j)
        T_r = T / T_c
        return R * T_c / P_c * (self.d_B0_d_Tr_expr(T_r) + w * self.d_B1_d_Tr_expr(T_r))

    def d_dij_d_Tr(self, i, j, T):
        r"""
        .. math::
            \frac{\partial \delta_{ij}}{\partial T_{\text{r}ij}} = 2\frac{\partial B_{ij}}{\partial T_{\text{r}ij}}-\frac{\partial B_{ii}}{\partial T_{\text{r}ij}}-\frac{\partial B_{jj}}{\partial T_{\text{r}ij}}
            :label: eq_d_dij

        .. todo::
            test this with symbolic differentiation of d_ik expression

        :param i: index for component *i*
        :param j: index for componen t *j*
        :param T: temperature [K]
        """
        if i == j:
            return 0.

        return 2. * self.d_Bij_d_Trij(i, j, T) \
               - self.d_Bij_d_Trij(i, i, T) \
               - self.d_Bij_d_Trij(j, j, T)

    def bar_HiR_RT(self, cas_k: str, ys: typing.List[typing.Union[float, typing.Any]], P, T):
        r"""Dimensionless residual partial molar enthalpy of component :math:`i`

        :param cas_k: cas number for component of interest
        :param ys: mole fractions
        :param P: pressure in Pa
        :param T: temperature in K
        """
        assert cas_k in self.cas_numbers, 'Cas number not found'
        k = self.cas_numbers.index(cas_k)
        T_c = self.T_cs[k]
        return self.ln_hat_phi_k_expr(k, ys, P, T) - P / R / T_c * (
                self.d_Bij_d_Trij(k, k, T)
                + 1. / 2. * sum(
            ys[i] * ys[j] * (2. * self.d_dij_d_Tr(i, k, T) - self.d_dij_d_Tr(i, j, T))
            for i in range(self.num_components) for j in range(self.num_components)
        )
        )

    def bar_HiR_star(self, T_star, cas_k: str, ys: typing.List[typing.Union[float, typing.Any]], P, T):
        r""" Returns the scaled entropy useful for computations :math:`\bar{H}_i^{\text{R},\star}`,
        as defined in Equation :eq:`barHiRstar_definition`

        :param cas_k: cas number of component *k*
        :param ys: mole fractions
        :param P: pressure in Pa
        :param T: temperature in K
        """
        return T_star * self.bar_HiR_RT(cas_k, ys, P, T)

    def bar_SiR_R(self, cas_k: str, ys: typing.List[typing.Union[float, typing.Any]], P, T):
        r"""Dimensionless residual partial molar entropy of component :math:`i`

        :param cas_k: cas number for component of interest
        :param ys: mole fractions
        :param P: pressure in Pa
        :param T: temperature in K
        :returns: Equation :eq:`barSiR`
        """
        assert cas_k in self.cas_numbers, 'Cas number not found'
        k = self.cas_numbers.index(cas_k)
        T_c = self.T_cs[k]
        return -P / R / T_c * (
                self.d_Bij_d_Trij(k, k, T)
                + sum(
            1. / 2. * ys[i] * ys[j] * (2. * self.d_dij_d_Tr(i, k, T) - self.d_dij_d_Tr(i, j, T))
            for i in range(self.num_components) for j in range(self.num_components)
        )
        )

    def bar_ViR_RT(self, cas_k: str, ys: typing.List[typing.Union[float, typing.Any]], P, T):
        r""" residual Partial molar volume for component *i*

        .. note::
            This expression does not depend on :math:`P`

        :param cas_k: cas number for component of interest
        :param ys: mole fractions
        :param P: pressure in Pa
        :param T: temperature in K
        :returns: Equation :eq:`barViR`
        """
        assert cas_k in self.cas_numbers, 'Cas number not found'
        k = self.cas_numbers.index(cas_k)
        return 1. / R / T(
            self.B_ij_expr(k, k, T)
            + 1. / 2. * sum(
                ys[i] * ys[j] * (2. * self.d_ik_expr(i, k, T) - self.d_ik_expr(i, j, T))
                for i in range(self.num_components) for j in range(self.num_components)
            )
        )

    def M_R_dimensionless(self, method: callable, ys: typing.List[typing.Union[float, typing.Any]], P: float, T: float):
        """Residual property of :math:`X` for mixture.

        Similar to Equation :eq:`residual_molar` but in dimensionless form

        :param method: function to compute partial molar property of compound
        :type method: callable
        """
        return sum(
            ys[i] * method(self.cas_numbers[i], ys, P, T) for i in range(self.num_components)
        )

    def S_R_R(self, *args):
        r"""Residual entropy of mixture :math:`S^\mathrm{R}/R`"""
        return self.M_R_dimensionless(self.bar_SiR_R, *args)

    def H_R_RT(self, *args):
        r"""Residual enthalpy of mixture :math:`H^\mathrm{R}/R/T`"""
        return self.M_R_dimensionless(self.bar_HiR_RT, *args)

    def G_R_RT(self, *args):
        r"""Residual free energy of mixture :math:`G^\mathrm{R}/R/T`"""
        return self.M_R_dimensionless(self.bar_GiR_RT, *args)

    def plot_residual_HSG(self, P, T, ax=None, fig=None) -> typing.Tuple[plt.figure, plt.subplot]:
        """Plot dimensionless residual properties as a function of mole fraction

        :param P: pressure in Pa
        :param T: Temperature in K
        :param ax: matplotlib ax, defaults to None
        :param fig: matplotlib figure, defautls to None
        """

        assert self.num_components == 2, 'Plotting only implemented for binary mixtures'

        if ax is None:
            if fig is None:
                fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.set_ylabel('Dimensionless')
            ax.set_xlabel('Gas-Phase Mole fraction %s' % self.compound_names[0])

        y_1 = np.linspace(0., 1.)
        HR_RT = np.array(list(self.H_R_RT([i, 1. - i], P, T) for i in y_1))
        GR_RT = np.array(list(self.G_R_RT([i, 1. - i], P, T) for i in y_1))
        SR_R = np.array(list(self.S_R_R([i, 1. - i], P, T) for i in y_1))
        ax.plot(y_1, HR_RT, '-', label='$H^\\mathrm{R}/(RT)$')
        ax.plot(y_1, GR_RT, '--', label='$G^\\mathrm{R}/(RT)$')
        ax.plot(y_1, SR_R, '-.', label='$S^\\mathrm{R}/R$')
        ax.set_title('Residual properties at T = %5.2f K and P = %4.3e Pa' % (T, P))
        ax.legend()
        return fig, ax



"""
Input
=====

Helper classes for inputting parameters
"""


class IdealMixture:
    r"""

    :param num_components: number of components
    :type num_components: int
    :param dippr_nos: dippr numbers of components
    :type dippr_nos: typing.Optional[typing.Union[str, None]]
    :param compound_names: names of components
    :type compound_names: typing.Optional[typing.Union[str, None]]
    :param cas_numbers: cas registry numbers
    :type cas_numbers:  typing.Optional[typing.Union[str, None]]
    """
    def __init__(self, **kwargs):
        self.num_components = kwargs.pop('num_components', 0)
        self.dippr_nos = kwargs.pop('dippr_nos', None)
        self.compound_names = kwargs.pop('compound_names', None)
        self.cas_numbers = kwargs.pop('cas_numbers', None)
        self._has_been_setup = False

    def setup(self):
        """setup input parameters"""
        self._find_number_components()
        self.check_input()
        self._update_params()
        self._has_been_setup = True

    def _find_number_components(self):
        if self.dippr_nos is not None:
            self.num_components = len(self.dippr_nos)
        if self.compound_names is not None:
            self.num_components = len(self.compound_names)
        if self.cas_numbers is not None:
            self.num_components = len(self.cas_numbers)

    def _update_params(self):
        if self.dippr_nos is None:
            self.dippr_nos = [None for I in range(self.num_components)]
        if self.compound_names is None:
            self.compound_names = [None for I in range(self.num_components)]
        if self.cas_numbers is None:
            self.cas_numbers = [None for I in range(self.num_components)]

    def check_input(self):
        assert self.num_components > 0, 'No components found!'

    def get_point_input(self, i: int) -> dict:
        """

        :param i: index of point
        :return: keyword arguments for point input
        """
        return {'dippr_no': self.dippr_nos[i] if self.dippr_nos is not None else None,
                'compound_name': self.compound_names[i] if self.compound_names is not None else None,
                'cas_number': self.cas_numbers[i] if self.cas_numbers is not None else None}

    def set_point_input(self, i: int, **kwargs):
        """

        :param i: index of point
        """
        self.dippr_nos[i] = kwargs.pop('dippr_no')
        self.compound_names[i] = kwargs.pop('compound_name')
        self.cas_numbers[i] = kwargs.pop('cas_number')

    def check_params(self):
        pass


class RealMixture(IdealMixture):
    r"""

    :param MWs: molecular weights of component sin g/mol
    :type MWs: typing.Optional[typing.List[float]]
    :param P_cs: critical pressures of componets in Pa
    :type P_cs: typing.Optional[typing.List[float]]
    :param T_cs: critical temperatures of pure components in K
    :type T_cs: typing.Optional[typing.List[float]]
    :param V_cs: critical molar volumes of pure components in m**3/mol
    :type V_cs: typing.Optional[typing.List[float]]
    :param ws: accentric factors of components
    :type ws: typing.Optional[typing.List[float]]
    :param k_ij: equation of state mixing rule in calculation of critical temperautre, see Equation :eq:`Tc_combine`. When :math:`i=j` and for chemical similar species, :math:`k_{ij}=0`. Otherwise, it is a small (usually) positive number evaluated from minimal :math:`PVT` data or, in the absence of data, set equal to zero.
    :type k_ij: typing.Union[float, typing.List[float]], defaults to 0

    """

    def __init__(self, **kwargs):
        IdealMixture.__init__(self, **kwargs)
        self.MWs = kwargs.pop('MWs', None)
        self.P_cs = kwargs.pop('P_cs', None)
        self.V_cs = kwargs.pop('V_cs', None)
        self.Z_cs = kwargs.pop('Z_cs', None)
        self.T_cs = kwargs.pop('T_cs', None)
        self.ws = kwargs.pop('ws', None)
        self.k_ij = kwargs.pop('k_ij', 0.)

    def _find_number_components(self):
        IdealMixture._find_number_components(self)
        if self.dippr_nos is None and self.compound_names is None and self.cas_numbers is None:
            assert (
                    self.MWs is not None
                    and self.P_cs is not None
                    and self.T_cs is not None
                    and self.V_cs is not None
                    and self.Z_cs is not None
                    and self.ws is not None

                    ), 'Incorrect input'
            assert self.num_components == 0, 'What happended?'
        if self.num_components == 0:
            self.num_components = len(self.MWs)

        if isinstance(self.k_ij, float):
            val = self.k_ij
            self.k_ij = [[0. for i in range(self.num_components)] for j in range(self.num_components)]
            for i in range(self.num_components):
                for j in range(self.num_components):
                    if i != j:
                        self.k_ij[i][j] = val

        for i in range(self.num_components):
            for j in range(self.num_components):
                assert abs(self.k_ij[i][j]) < 1e-8, 'K[i][i] must be zero!'

    def _update_params(self):
        IdealMixture._update_params(self)
        if self.MWs is None:
            self.MWs = [None for I in range(self.num_components)]
        if self.P_cs is None:
            self.P_cs = [None for I in range(self.num_components)]
        if self.Z_cs is None:
            self.Z_cs = [None for I in range(self.num_components)]
        if self.T_cs is None:
            self.T_cs = [None for I in range(self.num_components)]
        if self.V_cs is None:
            self.V_cs = [None for I in range(self.num_components)]
        if self.ws is None:
            self.ws = [None for I in range(self.num_components)]

    def get_point_input(self, i: int) -> dict:
        """

        :param i: index of point
        :return: keyword arguments for point input
        """
        data = IdealMixture.get_point_input(self, i)
        data['MW'] = self.MWs[i] if self.MWs is not None else None
        data['P_c'] = self.P_cs[i] if self.P_cs is not None else None
        data['V_c'] = self.V_cs[i] if self.V_cs is not None else None
        data['Z_c'] = self.Z_cs[i] if self.Z_cs is not None else None
        data['T_c'] = self.T_cs[i] if self.T_cs is not None else None
        data['w'] = self.ws[i] if self.ws is not None else None
        return data

    def set_point_input(self, i: int, **kwargs):
        """

        :param i: index of point
        """
        IdealMixture.set_point_input(self, i, **kwargs)
        self.MWs[i] = kwargs.pop('MW')
        self.P_cs[i] = kwargs.pop('P_c')
        self.T_cs[i] = kwargs.pop('T_c')
        self.V_cs[i] = kwargs.pop('V_c')
        self.Z_cs[i] = kwargs.pop('Z_c')
        self.ws[i] = kwargs.pop('w')

    def check_params(self):
        IdealMixture.check_params(self)
        if len(set(self.T_cs)) > 1:
            assert len(set(self.cas_numbers)) > 1, 'Errors anticipated when cas numbers equal if other props are not'


r"""

Theory
------

The generic cubic equation of state is :cite:`Perry`

.. math::
    P = \frac{RT}{V - b} - \frac{a(T)}{(V + \epsilon b)(V + \sigma b)}

where :math:`\epsilon` and :math:`\sigma` are pure numbers (the same for all substances),
and :math:`a(T)` and :math:`b` are given by the following equations

.. math::
    a(T) = \Psi \frac{\alpha(T_r)R^2T_{\text{c}}^2}{P_{\text{c}}}
    :label: a_expr

.. math::
    b = \Omega\frac{RT_\text{c}}{P_\text{c}}

The compressibility factor can be calculated by solving the following equation


.. math::
    Z - \left(1 + \beta - \frac{q\beta (Z - \beta)}{(Z + \epsilon\beta)(Z + \sigma\beta)}\right) = 0
    :label: eq_cubic_residual

where

.. math::
    \beta = \Omega\frac{P_\text{r}}{T_\text{r}}
    :label: beta_expr

and

.. math::
    q = \frac{\Psi \alpha(T_\text{r})}{\Omega T_\text{r}}
    :label: q_expr


An iterative routine to calculate :math:`Z` using Equation :eq:`eq_cubic_residual` is implemented,
following :cite:`Perry`, where

.. math::
    Z^\text{new} = 1 + \beta - \frac{q\beta (Z^\text{old} - \beta)}{(Z^\text{old} + \epsilon\beta)(Z^\text{old} + \sigma\beta)}
    :label: eq_iteration

that is continued until the following is true

.. math::
    \|\frac{Z^\text{new} - Z^\text{old}}{Z^\text{new} + Z^\text{old}}\times 200\| < \text{tol}
    :label: tolerance

Residual Molar Properties
-------------------------

.. math::
    \frac{H^\text{R}}{RT} = Z - 1 + \left[\frac{\text{d} \ln \alpha(T_\text{r})}{\text{d} \ln T_\text{r}} - 1\right]qI
    :label: H_R_RT_cubic

.. math::
    \frac{G^\text{R}}{RT} = Z - 1 + \ln(Z-\beta) - qI
    :label: G_R_RT_cubic

.. math::
    \frac{S^\text{R}}{R} = \ln(Z-\beta) + \frac{\text{d} \ln \alpha(T_\text{r})}{\text{d} \ln T_\text{r}} qI
    :label: S_R_R_cubic

where the following functions have been defined

.. math::
    I = \frac{1}{\sigma - \epsilon}\ln{\left(
        \frac{Z + \sigma \beta}{Z + \epsilon \beta}
    \right)}
    :label: I_expr

.. math::
    1 + \beta - \frac{q\beta (Z - \beta)}{(Z + \epsilon\beta)(Z + \sigma\beta)}
    :label: eq_cubic_rhs


Fugacity Coefficients
---------------------
The pure-component fugacity coefficient is *defined* as

.. math::
    \frac{G_i^\text{R}}{RT} = \ln\hat{\phi}_i

So that, from Equation :eq:`G_R_RT_cubic`,

.. math::
    \ln\hat{\phi}_i = Z_i - 1 + \ln(Z_i-\beta_i) - q_iI_i
    :label: ln_hat_phi_i


Mixtures
--------
.. warning::
    Not implemented yet!

.. todo::
    Implement mixtures with cubic equations of state

"""

import matplotlib.pyplot as plt
import numpy as np
from chem_util.chem_constants import gas_constant as R
from chem_util.math import percent_difference

from ..critical_constants import CriticalConstants


class Cubic(CriticalConstants):
    """Generic Cubic Equation of State

    :param sigma: Parameter defined by specific equation of state, :math:`\\sigma`
    :param epsilon: Parameter defined by specific equation of state, :math:`\\epsilon`
    :param Omega: Parameter defined by specific equation of state, :math:`\\Omega`
    :param Psi: Parameter defined by specific equation of state, :math:`\\Psi`
    :param tol: tolerance for iteration (see Equation :eq:`tolerance`), set to 0.01
    :param log: function for computing natural log, defaults to :code:`numpy.log`
    :param exp: function for computing exponential, defaults to :code:`numpy.exp`

    """

    def __init__(self, sigma: float, epsilon: float, Omega: float, Psi: float,
                 dippr_no: str = None, compound_name: str = None, cas_number: str = None,
                 log: callable = np.log, exp: callable = np.exp, name: str = 'cubic',
                 **kwargs):
        """

        :param kwargs: for customized critica constants
        """
        CriticalConstants.__init__(self, dippr_no, compound_name, cas_number, **kwargs)
        self.sigma = sigma
        self.epsilon = epsilon
        self.Omega = Omega
        self.Psi = Psi
        self.b = self.Omega * R * self.T_c / self.P_c
        self.tol = 0.01
        self.log = log
        self.exp = exp
        self.name = name

    """Expressions"""

    def alpha_expr(self, T_r):
        """An empirical expression, specific to a particular form of the equation of state

        :param T_r: reduced temperature (T/Tc), dimensionless
        :return: :math:`\\alpha (T_r)` see Table ??
        """
        raise NotImplementedError

    def beta_expr(self, T, P):
        """

        :param T:  temperature in K
        :param P: pressure in Pa
        :return: :math:`\\beta` (see Equation :eq:`beta_expr`)
        """
        P_r = P / self.P_c
        T_r = T / self.T_c
        return self.Omega * P_r / T_r

    def q_expr(self, T):
        """


        :param T: temperautre in K
        :returns: :math:`q` (see Equation :eq:`q_expr`)
        """
        T_r = T / self.T_c
        return self.Psi * self.alpha_expr(T_r) / self.Omega / T_r

    def a_expr(self, T):
        """

        :param T: temperautre in K
        :returns: :math:`a(T)` (see Equation :eq:`a_expr`)
        """
        return self.Psi * self.alpha_expr(T / self.T_c) * R * R * self.T_c * self.T_c / self.P_c

    def Z_expr(self, P, V, T):
        return P * V / R / T

    def I_expr(self, P, V, T):
        """

        :param P: pressure in Pa
        :param V: molar volume in m**3/mol
        :param T: temperature in K
        :returns: :math:`I` (see Equation :eq:`I_expr`)
        """
        Z = self.Z_expr(P, V, T)
        B = self.beta_expr(T, P)
        return self.log(
            (Z + self.sigma * B) / (Z + self.epsilon * B)
        ) / (self.sigma - self.epsilon)

    def d_ln_alpha_d_ln_Tr(self, T_r):
        """

        :param T_r: reduced temperature [dimensionless]
        :return: Expression for :math:`\\frac{\\mathrm{d} \\ln \\alpha(T_\\mathrm{r})}{\\mathrm{d} \\ln T_\\mathrm{r}}`
        """
        raise NotImplementedError

    def H_R_RT_expr(self, P, V, T):
        r"""Dimensionless residual enthalpy


        :param P: pressure in Pa
        :param V: molar volume in m**3/mol
        :param T: temperature in K
        :returns: :math:`\frac{H^\text{R}}{RT}`, see Equation :eq:`H_R_RT_cubic`
        """
        return self.Z_expr(P, V, T) - 1 \
               + (self.d_ln_alpha_d_ln_Tr(T / self.T_c) - 1) * self.q_expr(T) * self.I_expr(P, V, T)

    def G_R_RT_expr(self, P, V, T):
        r"""Dimensionless residual gibbs

        :param P: pressure in Pa
        :param V: molar volume in m**3/mol
        :param T: temperature in K
        :returns: :math:`\frac{G^\text{R}}{RT}`, see Equation :eq:`G_R_RT_cubic`
        """
        Z = self.Z_expr(P, V, T)
        return Z - 1 - self.log(Z - self.beta_expr(T, P)) - self.q_expr(T) * self.I_expr(P, V, T)

    def S_R_R_expr(self, P, V, T):
        r"""Dimensionless residual entropy

        :param P: pressure in Pa
        :param V: molar volume in m**3/mol
        :param T: temperature in K
        :returns: :math:`\frac{S^\text{R}}{R}`, see Equation :eq:`S_R_R_cubic`
        """
        Z = self.Z_expr(P, V, T)
        return self.log(Z - self.beta_expr(T, P)) \
               + self.d_ln_alpha_d_ln_Tr(T / self.T_c) * self.q_expr(T) * self.I_expr(P, V, T)

    def ln_hat_phi_k_expr(self, P, V, T):
        r"""

        :param P: pressure in Pa
        :param T: temperature in K
        :param V: molar volume in m**3/mol
        :returns: :math:`\ln{\hat{\phi}_k`, see Equation :math:`ln_hat_phi_i`
        """
        return self.G_R_RT_expr(P, V, T)

    def hat_phi_i_expr(self, *args):
        r"""expression for fugacity coefficient
        :returns: :math:`\exp\left({\ln{\hat{\phi}_i}}\right)`
        """
        return self.exp(self.ln_hat_phi_k_expr(*args))

    """Solving equations"""

    def coefficients(self, T, P):
        """Polynomial oefficients for cubic equation of state

        .. math::
            Z^3 c_0 + Z^2c_1 + Z*c_2 + c_3 = 0

        :return: :code:`(c_0, c_1, c_2, c_3)`
        """
        B = self.beta_expr(T, P)
        e = self.epsilon
        o = self.sigma
        q = self.q_expr(T)
        return [
            1,
            B * (o + e - 1) - 1,
            B * (B * (e * o - e - o) - e - o + q),
            -B * B * (e * o * (1 - B) - q)
        ]

    def cardano_constants(self, T, P):
        """

        :param T: temperature [T]
        :param P: pressure [Pa]
        :return: cardano constants p, q
        :rtype: tuple
        """
        _, b_2, b_1, b_0 = self.coefficients(T, P)
        return (
            b_1 - b_2 * b_2 / 3,  # p
            b_0 - b_1 * b_2 / 3. + 2. * b_2 * b_2 * b_2 / 27.,  # q
            b_2
        )

    def one_root(self, T, P):
        p, q, b_2 = self.cardano_constants(T, P)
        d = (p / 3) * (p / 3) * (p / 3) + (q / 2) * (q / 2)
        return (-q / 2. + d ** 0.5) ** (1. / 3.) + (-q / 2. - d ** 0.5) ** (1. / 3.) - b_2 / 3.

    def num_roots(self, T, P):
        """Find number of roots

        See :cite:`Loperena2012,Deiters2002`

        :param T: temperature in K
        :param P: pressure in Pa
        :return: number of roots
        """
        p, q, _ = self.cardano_constants(T, P)
        discriminant = (p / 3) * (p / 3) * (p / 3) + (q / 2) * (q / 2)
        if discriminant > 0:
            return 1

        return 3

    def print_roots(self, T, P):
        """Check to see if all conditions have one root"""
        print('%s with %s EOS at T %5.1f K, P %4.2f MPa has %i roots' % (
            self.compound_name, self.name, T, P * 1e-6, self.num_roots(T, P)))

    def Z_right_hand_side(self, Z, beta, q):
        """
        Estimate of compressibility of vapor :cite:`Perry`, used for iterative methods

        :return: RHS of residual, see Equation :eq:`eq_cubic_rhs`
        """
        return 1 + beta - q * beta * (Z - beta) / (Z + self.epsilon * beta) / (Z + self.sigma * beta)

    def residual(self, P, V, T):
        """

        :param P: pressure in Pa
        :param V: volume in [mol/m**3]
        :param T: temperature in K
        :return: residual for cubic equation of state (Equation :eq:`eq_cubic_residual`)
        """
        Z = self.Z_expr(P, V, T)
        return Z - self.Z_right_hand_side(Z, self.beta_expr(T, P), self.q_expr(T))

    def iterate_to_solve_Z(self, T, P) -> float:
        """
        Iterate to compute :math:`Z` using Equation :eq:`eq_iteration`
        util termination condition (Equation :eq:`tolerance` is met)

        :param T: temperature in K
        :param P: pressure in Pa
        :return: compressibility factor
        """
        beta = self.beta_expr(T, P)
        q = self.q_expr(T)
        Z_nm1 = 1.  # initial guess
        Z_n = self.Z_right_hand_side(Z_nm1, beta, q)
        while abs(percent_difference(Z_n, Z_nm1)) > self.tol:
            Z_nm1 = Z_n
            Z_n = self.Z_right_hand_side(Z_nm1, beta, q)

        return Z_n

    def plot_Z_vs_P(self, T: float, P_min: float, P_max: float, symbol='o', ax: plt.axis = None, fig: plt.figure = None,
                    **kwargs):
        """Plot compressibility as a function of pressure

        :param T: temperature [K]
        :param P_min: minimum pressure for plotting [Pa]
        :param P_max: maximum pressure for plotting [Pa]
        :param symbol: marker symbol, defaults to 'o'
        :param ax: matplotlib axes for plotting, defaults to None
        :param kwargs: keyword arguments for plotting
        """
        if ax is None:
            if fig is None:
                fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.set_ylabel('$Z$')
            ax.set_xlabel('$P$ [Pa]')

        P = np.linspace(P_min, P_max)
        Z = [self.iterate_to_solve_Z(T, pressure) for pressure in P]
        ax.plot(P, Z, symbol, markerfacecolor='None', **kwargs)
        return fig, ax


class RedlichKwong(Cubic):
    r"""Redlich-Kwong Equation of State :cite:`RK`

    This Equation of state has the following parameters :cite:`Smith2005`

    ===========================       ==============================
            Symbol                                  Value
    ===========================       ==============================
    :math:`\alpha(T_\text{r})`          :math:`1/\sqrt{T_\text{r}}`
    :math:`\sigma`                               1
    :math:`\epsilon`                             0
    :math:`\Omega`                                  0.08664
    :math:`\Psi`                                    0.42748
    ===========================       ==============================


    >>> from realgas.eos.cubic import RedlichKwong
    >>> model = RedlichKwong(compound_name='Propane')
    >>> model.sigma
    1
    >>> model.epsilon
    0
    >>> model.Omega
    0.08664
    >>> model.Psi
    0.42748

    """

    def __init__(self, **kwargs):
        Cubic.__init__(self, sigma=1, epsilon=0, Omega=0.08664, Psi=0.42748, name='RK', **kwargs)

    def alpha_expr(self, T_r):
        return 1 / T_r ** 0.5

    def d_ln_alpha_d_ln_Tr(self, T_r):
        return -1. / 2. / T_r / T_r


class SoaveRedlichKwong(RedlichKwong):
    r"""Soave Redlich-Kwong Equation of State :cite:`SRK`

    This equation of state has the following parameters

    ===========================       ====================================================
            Symbol                                  Value
    ===========================       ====================================================
    :math:`\alpha(T_\text{r})`         :math:`\left[1 + f_w(1-\sqrt{T_\text{r}})\right]^2`
    :math:`\sigma`                               :math:`1`
    :math:`\epsilon`                             :math:`0`
    :math:`\Omega`                                :math:`0.08664`
    :math:`\Psi`                                    :math:`0.42748`
    ===========================       ====================================================

    where

    .. math::
        f_w = 0.480 + 1.574\omega - 0.176\omega^2
        :label: f_w_SRK

    >>> from realgas.eos.cubic import SoaveRedlichKwong
    >>> model = SoaveRedlichKwong(compound_name='Water')
    >>> model.Omega
    0.08664
    >>> model.sigma
    1
    >>> model.epsilon
    0
    >>> model.Psi
    0.42748
    >>> model.f_w_rule(0.)
    0.48
    >>> model.f_w_rule(1.)
    1.878

    :param f_w: empirical expression used in :math:`\\alpha` [dimensionless], see Equation :eq:`f_w_SRK`
    :type f_w: float, derived
    """

    def __init__(self, **kwargs):
        Cubic.__init__(self, sigma=1, epsilon=0, Omega=0.08664, Psi=0.42748, name='SRK', **kwargs)
        self.f_w = self.f_w_rule(self.w)

    def f_w_rule(self, w):
        r"""

        :param w: accentric factor
        :returns: :math:`f_w`, see Equation :eq:`f_w_SRK`
        """
        return 0.480 + 1.574 * w - 0.176 * w * w

    def alpha_expr(self, T_r):
        val = 1. + self.f_w * (1 - T_r ** 0.5)
        return val * val

    def d_ln_alpha_d_ln_Tr(self, T_r):
        T_r_sqrt = T_r ** 0.5
        return (
                self.f_w / (self.f_w * (T_r_sqrt - 1) - 1) / T_r
        )


class PengRobinson(SoaveRedlichKwong):
    r"""Peng-Robinson Equation of State :cite:`peng-robinson`

    This equation of state has the following parameters

    ===========================       ====================================================
            Symbol                                  Value
    ===========================       ====================================================
    :math:`\alpha(T_\text{r})`         :math:`\left[1 + f_w(1-\sqrt{T_\text{r}})\right]^2`
    :math:`\sigma`                               :math:`1 + \sqrt{2}`
    :math:`\epsilon`                             :math:`1 - \sqrt{2}`
    :math:`\Omega`                                :math:`0.07780`
    :math:`\Psi`                                    :math:`0.45724`
    ===========================       ====================================================

    where

    .. math::
        f_w = 0.480 + 1.574\omega - 0.176\omega^2
        :label: f_w_PR

    >>> from realgas.eos.cubic import PengRobinson
    >>> model = PengRobinson(compound_name='Water')
    >>> model.Omega
    0.0778
    >>> model.sigma
    2.4142135
    >>> model.epsilon
    -0.414213
    >>> model.Psi
    0.45724
    >>> model.f_w_rule(0.)
    0.37464
    >>> model.f_w_rule(1.)
    1.64698

    :param f_w: empirical expression used in :math:`\\alpha` [dimensionless?]
    :type f_w: float, derived
    """

    def __init__(self, **kwargs):
        Cubic.__init__(self, sigma=1 + np.sqrt(2), epsilon=1 - np.sqrt(2), Omega=0.07780, Psi=0.45724, name='PR',
                       **kwargs)
        self.f_w = self.f_w_rule(self.w)

    def f_w_rule(self, w):
        r"""

        :param w: accentric factor
        :returns: :math:`f_w`, see Equation :eq:`f_w_PR`
        """
        return 0.37464 + 1.54226 * w - 0.26992 * w * w



r"""

We **define** partial molar property :math:`\bar{M}_i` of species *i* in a mixture as

.. math::
    \bar{M}_i = \left(\frac{\partial(nM)}{\partial n_i}\right)_{P, T, n_j}
    :label: partial_molar_property

The mixture property is related to the partial molar property as

.. math::
    nM=\sum_i n_i\bar{M}_i

or, in terms of gas-phase mole fractions :math:`y_i`,

.. math::
    M=\sum_i y_i\bar{M}_i

The following relationships also hold

.. math::
    \bar{M}_i = \bar{M}^\text{IG} + \bar{M}^\text{R}
    :label: residual_partial_molar

.. math::
    M^\text{R} = \sum_i y_i\bar{M}^\text{R}
    :label: residual_molar


Ideal Gas
---------

The *Gibbs theorem* is

    A partial molar property (other than volume) of a contituent species
    in an ideal-gas mixture is equal to the corresponding molar property
    of the species as a pure ideal gas at the mixture temperature but
    at a pressure equal to its partial pressure in the mixture


The partial molar volume of an ideal gas, :math:`\bar{V}_i^\text{IG}`, is

.. math::
    \bar{V}_i^\text{IG} = \frac{RT}{P}
    :label: barViIG

The partial molar enthalpy of an ideal gas, :math:`\bar{H}_i^\text{IG}`, is

.. math::
    \bar{H}_i^\text{IG} = H_i^\text{IG}

which results from the enthalpy of an ideal gas being independent of pressure.
Therefore, we can compute the ideal gas partial molar enthalpy if we have
the ideal gas heat capacities, as follows

.. math::
    \bar{H}_i^\text{IG} = H_i^\text{IG} = \int_{T_\text{ref}}^T C_{\text{p},i}^\text{IG}\mathrm{d}T^\prime
    :label: barHiIG

where :math:`T_\text{ref}` is a reference temperature and :math:`T^\prime` is a dummy variable for integration.
Often times, we want to compute these quantities in dimensionless units,

.. math::
    \begin{align}
        \bar{H}_i^{\text{IG},\star} &= \frac{\bar{H}_i^\text{IG}}{RT_\text{ref}} \\
                                    &= \int_{1}^{T^\star} \frac{C_{\text{p},i}^\text{IG}}{R}\mathrm{d}T^\prime \\
                                    &=
    \end{align}

or

.. math::
    \bar{H}_i^{\text{IG},\star} = \int_{1}^{T^\star} C_{\text{p},i}^{\text{IG},\star}\mathrm{d}T^\prime
    :label: barHiIGstar

where

.. math::
    T^\star=T/T_\text{ref}

is a dimensionless variable
and

.. math::
    C_{\text{p},i}^{\text{IG},\star} = C_{\text{p},i}^\text{IG}/R

is a dimensionless parameter that is a function of :math:`T^\star`.

We note that the thermodynamic integration reference temperature does not have to be the same
as the temperature for scaling, but we have made them the same here for simplicity.

.. todo::
    Implement ideal gas partial molar entropy??
    Implement ideal gas partial molar free energy??
    This might not be useful though because these values seem to *depend* on mixture properties

Residual
--------

The residual partial molar volume of component *i*, :math:`\bar{V}_i^\text{R}`,
can be calculated as

.. math::
    \bar{V}_i^\mathrm{R} = RT\left(\frac{\partial \ln\hat{\phi}_i}{\partial P}\right)_{T,y}\\
    :label: barViR_definition

Defining the dimensionless quantity

.. math::
    \bar{V}_i^{\mathrm{R},\star} = \frac{\bar{V}_i^\text{R}}{RT_\text{ref}}

The expression in dimensionless units can be simplified to

.. math::
    \bar{V}_i^{\mathrm{R},\star} = T^\star\left(\frac{\partial \ln{\hat{\phi}_i}}{\partial P}\right)_{T^\star, y}
    :label: barViRstar_definition

The residual partial molar enthalpy of component *i*, :math:`\bar{H}_i^\text{R}`, can be calculated as

.. math::
    \bar{H}_i^{\text{R}} = -RT^2\left(\frac{\partial \ln\hat{\phi}_i}{\partial T}\right)
    :label: barHiR_definition

Defining the dimensionless quantity

.. math::
    \begin{align}
        \bar{H}_i^{\text{R},\star} &= \frac{\bar{H}_i^\text{R}}{RT_\text{ref}} \\
         &= -\frac{RT^2}{RT_\text{ref}}\left(\frac{\partial \ln\hat{\phi}_i}{\partial T}\right) \\
         &= -\frac{RT^2}{RT_\text{ref}^2}\left(\frac{\partial \ln\hat{\phi}_i}{\partial T^\star}\right) \\
    \end{align}

The expression in dimensionless units is computed as

.. math::
    \bar{H}_i^{\text{R},\star} = -\left(T^\star\right)^2\left(\frac{\partial \ln\hat{\phi}_i}{\partial T^\star}\right) \\
    :label: barHiRstar_definition

The residual partial molar free energy of component *i*, :math:`\bar{G}_i^\text{R}`,
which *defines* the fugacity coefficient :math:`\hat{\phi}_i`, is

.. math::
    \bar{G}_i^\text{R} = RT \ln\hat{\phi}_i
    :label: barGiR_definition


With these definitions, however, we note that we need an equation of state to calculate the partial molar properties.
In this package, the second-order virial equation of state currently implements the necessary derivatives.

"""


from realgas.eos.virial import SecondVirialMixture
import realgas.input as inp
from realgas.cp import CpIdealGas, CpStar
from chem_util.chem_constants import gas_constant as R
from realgas.input import IdealMixture
import numpy as np
import typing


class Mixture(SecondVirialMixture):
    """
    :param ideal: whether or not ideal gas, defaults to True
    :type ideal: bool, optional
    :param kwargs: key-word arguments for :class:`realgas.eos.virial.SecondVirialMixture`


    """

    def __init__(self, cp_args: typing.List[dict], ideal=True, **kwargs):
        self.ideal = ideal
        self.data = [
            CpIdealGas(**i) for i in cp_args
        ]
        self.cas_numbers = [I.cas_number for I in self.data]
        if not self.ideal:
            SecondVirialMixture.__init__(self, **kwargs)

    def bar_Hi_IG(self, cas_i: str, T, T_ref=0):
        """

        :param T_ref: reference temperature in K for enthalpy, defaults to 0
        :param T: temperautre in K
        :return: :math:`\bar{H}_i^\text{IG}`, see Equation :eq:`barHiIG`
        """
        assert cas_i in self.cas_numbers, 'Cas number not found!'
        i = self.cas_numbers.index(cas_i)
        return self.data[i].cp_integral(T_ref, T)

    def bar_Vi_IG(self, T, P):
        """

        :param T: temperature in K
        :param P: pressure in Pa
        :return: :math:`\bar{V}_i^\text{IG}`, see Equation :eq:`barViIG`
        """
        return R * T / P

    def bar_Hi_R(self, cas_k: str, ys: typing.List[typing.Union[float, typing.Any]], P, T):
        assert not self.ideal, 'No residual props for ideal gas'
        return R * T * self.bar_HiR_RT(cas_k, ys, P, T)

    def bar_Vi_R(self, cas_k: str, ys: typing.List[typing.Union[float, typing.Any]], P, T):
        assert not self.ideal, 'No residual props for ideal gas'
        return R * T * self.bar_ViR_RT(cas_k, ys, P, T)

    def bar_Hi(self, cas_k: str, ys: typing.List[typing.Union[float, typing.Any]], P, T, T_ref=0):
        if self.ideal:
            return self.bar_Hi_IG(cas_k, T, T_ref=T_ref)

        return self.bar_Hi_IG(cas_k, T, T_ref=T_ref) + R * T * self.bar_HiR_RT(cas_k, ys, P, T)

    def bar_Vi(self, cas_k: str, ys: typing.List[typing.Union[float, typing.Any]], P, T):
        if self.ideal:
            return self.bar_Vi_IG(T, P)

        return self.bar_Vi_IG(T, P) + R * T * self.bar_Vi_R(cas_k, ys, P, T)

    def enthalpy(self, ys: typing.List[typing.Union[float, typing.Any]], P: float, T: float, T_ref=0.):
        """Residual property of :math:`X` for mixture.

        Similar to Equation :eq:`residual_molar` but in dimensionless form

        :param method: function to compute partial molar property of compound
        :type method: callable
        """
        return sum(
            ys[i] * self.bar_Hi(self.cas_numbers[i], ys, P, T, T_ref) for i in range(self.num_components)
        )


class MixtureDimensionless(SecondVirialMixture):
    """

    .. todo::
        add docs!

    :param ideal: whether or not ideal gas, defaults to True
    :type ideal: bool, optional
    :param kwargs: key-word arguments for :class:`realgas.eos.virial.SecondVirialMixture`
    """

    def __init__(self, cp_args: typing.List[dict], ideal=True, **kwargs):
        self.ideal = ideal
        self.data = [
            CpStar(**i) for i in cp_args
        ]
        self.cas_numbers = [I.cas_number for I in self.data]
        if not self.ideal:
            SecondVirialMixture.__init__(self, **kwargs)

    def bar_Hi_IG_star(self, cas_i, T_star, T_ref_star=0):
        """

        :param T_ref_star: dimensionless reference temperature in K for enthalpy, defaults to 0
        :param T_star: dimensionless temperature
        :return: :math:`\bar{H}_i^\text{IG}`, see Equation :eq:`barHiIGstar`
        """
        assert cas_i in self.cas_numbers, 'Cas number not found!'
        i = self.cas_numbers.index(cas_i)
        return self.data[i].cp_integral(T_star, T_ref_star)

    def bar_Hi_star(self, T_star, cas_k: str, ys: typing.List[typing.Union[float, typing.Any]], P, T):
        if self.ideal:
            return self.bar_Hi_IG_star(cas_k, T_star)

        return self.bar_Hi_IG_star(cas_k, T_star) + self.bar_HiR_star(T_star, cas_k, ys, P, T)

    def enthalpy_star(self, ys: typing.List[typing.Union[float, typing.Any]], P: float, T: float):
        """Residual property of :math:`X` for mixture.

        Similar to Equation :eq:`residual_molar` but in dimensionless form

        :param method: function to compute partial molar property of compound
        :type method: callable
        """
        return sum(
            ys[i] * self.bar_Hi_star(self.cas_numbers[i], ys, P, T) for i in range(self.num_components)
        )
